// We include a librairy to use a servomotor
#include <Servo.h>


/**************************************************
 *                                                *
 *                  Variables                     *
 *                                                *
 **************************************************/
 
#define samp_siz 6 // Number of values used to determine a frequency
#define rise_threshold 5 // Number of successive values upper than the ones before to consider that the heart beat magnitude is rising
Servo servo; // Variable associated with the servomotor
int time_buzzer=1; // Variable that determines the time where the buzzer is on and off
float reads[samp_siz], sum=0; // Table of the values got by the sensor and the sum of it
long int now, ptr; // now is th time in milli second and ptr is the index of 'reads' that we reaplace by the new value
float last=0, reader=0, start=0;
float first=0, second=0, third=0, before=0;
float first_freq=0, second_freq=0, third_freq=0; // To have a smoothed curve, we keep the last three frequencies
float frequency=0;
float frequency_threshold = 500;
int beat_number = 0; // Number of heart beats since the launch of the program.
float usual_heart_rate = 0;
bool rising;
int rise_count=0;
long int last_beat=0;
bool end_training = false;
int duration =25;

// Pin definition
const int BUZZER_PIN = 8; // Digital Pin
const int SERVO_PIN = 9; // Digital Pin
const int SENSOR_PIN = 0; // Analog Pin


/**************************************************
 *                                                *
 *                  Functions                     *
 *                                                *
 **************************************************/

void bip(){
  // This function uses the buzzer to produce a sound.
  for (int i=0; i<100;i++){
    digitalWrite(BUZZER_PIN, HIGH);
    delay(time_buzzer);
    digitalWrite(BUZZER_PIN, LOW);
    delay(time_buzzer);
  }
}

void hardware_connect() {
  // This function sets up the connections with the hardware.
  
  // Setting up the servomotor.
  servo.attach(SERVO_PIN);
  // Setting up the buzzer.
  pinMode(BUZZER_PIN, OUTPUT);
  digitalWrite(BUZZER_PIN, LOW);
  // Setting up the connection between the Arduino and the computer.
  Serial.begin(9600);
}

void init_heart_rate_calculation() {
  // This function initializes the variables needed to calculate the heart rate.
  for (int i=0 ; i<samp_siz ; i++)
    reads[i] = 0;
  sum = 0;
  ptr = 0;
}

void get_heart_mag(int duration) {
  // This function get an average of the magnitude of heart beat during 'duration' milli seconds.
  // It puts this value in the table that stocks the last 4 magnitudes (reads) and computes the average of these magnitudes.

  int n = 0; // counter of measures taken to compute the average
  start = millis(); // To compute the frequency
  reader = 0.;
  do {
    reader += analogRead(SENSOR_PIN); // Gets the value of the sensor
    n++;
    now = millis();
  } while (now < start + duration);
    
    
  reader /= n;  // average
  sum -= reads[ptr];
  sum += reader;
  reads[ptr] = reader;
  last = sum / samp_siz;
}

void get_frequency() {
  // This function measures the actual heart rate and update the variables related to it.
  first = millis() - last_beat;
  last_beat = millis();

  third_freq = second_freq;
  second_freq = first_freq;
  first_freq = 60000. / (0.4 * first + 0.3 * second + 0.3 * third);
  
  // To have a smoothed curve, we calculate the average of the last three frequencies.
  frequency = (third_freq+second_freq+first_freq)/3;
  Serial.print(frequency_threshold);
  Serial.print(",");
  Serial.println(frequency);
}

float threshold_by_percent(float moyenne) {
  // This function computes the threshold that determines the acceleration of the heart rate.
  float percent = 0.7; // This is the percentage of the mean that is consider as the threshold, you can have a threshold upper if you are athletic
  return (1+percent)*moyenne;
}

void frequency_threshold_setup() {
  // This function initializes the frequency's threshold that will stop the program.
  
  // The first frequency's values are false, so we calculate the usual heart rate of the person from the 20th measure.
  // We calculate then the average of 30 frequencies to have the usual heart rate.
  
  beat_number += 1;
  if (beat_number > 10 and beat_number < 25)
    usual_heart_rate += frequency;
  else if (beat_number==25) {
    usual_heart_rate = usual_heart_rate/14;
    frequency_threshold=threshold_by_percent(usual_heart_rate);
    bip();
  }
}

bool acceleration(float heart_rate, float threshold) {
  // This function is used to detect an acceleration of the heart rate.
  // The acceleration is defined by a threshold of heart rate.
  return heart_rate > threshold;
}

void rotate_servo(int from, int to, int del) {
  // This function rotates the servomotor from the angle 'from' to the angle 'to'.
  // The servo rotates of 1° every 'del' millisecond.
  if (from < to) {
    // Control of the direction of rotation
    for (int angle=from ; angle<=to ; angle++) {
      servo.write(angle);
      delay(del);
    }
  } else {
    for (int angle=from ; angle>=to ; angle--) {
      servo.write(angle);
      delay(del);
    }
  }
}

void not_enough() {
  // This function mooves the servomotor with little movements.
  // The rotation is made from 45° to 135° with a delay of 2s between each angle.
  rotate_servo(90, 135, 2); // The rotation begins at 90°
  rotate_servo(135, 45, 2);
  rotate_servo(45, 90, 2);// The rotation ends at 90°
}

void enough() {
  // When the user did enough sport, the coach stops rotating the head and the buzzer bips.
  bip();
  end_training = true;
}

void action() {
  // This function triggers the action to do depending on the heart rate.
  if (acceleration(frequency,frequency_threshold))
    // The heart rate is upper than the threshold, the user did enough sport.
    enough();
  else
    // The user did not enough sport.
    not_enough();
}

void heart_beat_rising() {
  // This function updates the frequency's measure and trigger the action to do.
  
  rising = true;
  
  get_frequency();
  
  // We define an frequency threshold at each launch of the program.
  // Above this threshold, we consider that you made enough sport.
  frequency_threshold_setup();
  
  action(); // The frequency triggers an action.
        
  third = second;
  second = first;
}

void training() {
  // This function defines the actions to do while the user did not enough sport.
  get_heart_mag(25);
  
  if (last > before) {
    // In this case, the average magnitude of heartbeat is bigger than the one before.
    rise_count++; // The number of value during the rise of magnitude is increased by one.
    // This variable enables a control whether it is really a rise or just a lonely value.
  
    if (!rising && rise_count > rise_threshold)
      // In this case, we consider that the heart beat magnitude is rising.
      heart_beat_rising();
  } else {
    rising = false;
    rise_count = 0;
  }
  before = last;
  ptr++;
  ptr %= samp_siz;
}


/**************************************************
 *                                                *
 *                Main functions                  *
 *                                                *
 **************************************************/
 
void setup() {
  // Setting up the hardware connection.
  hardware_connect();

  // Setting up of the variables that help to calculate the heart rate.
  init_heart_rate_calculation();
}

void loop() {
  if (!end_training)
    // The training, is not finished, we continue the program.
    training();
}
